// version 1.

// Library in use
#include <Wire.h>      //library for I2C protocol
#include <PWM.h>     //library to setup PWM frequency and pulse width
#include <LiquidCrystal_I2C.h>  // library to drive the LCD display
#include <SPI.h>         //library for SPI  protocol
#include <NRFLite.h>  //  library to drive the Wireless communication module NRF24L01
#include <EEPROM.h>  // library to read or write data to the internal EEPROM

// List of subroutines
// void pw_adj()  pulse width adjustment
// void meas()  Do measurements
// void config_ads1115a  setup  ADS1115 A and start a single measurement
// void config_ads1115b  setup  ADS1115 B and start a single measurement
// int v_ads1115a() setup ADS1115 A to send measured data
// int v_ads1115b() setup ADS1115 A to send measured data
// int reg_ads1115a reading ADS1115 A measurement value
// int reg_ads1115b reading ADS1115 B measurement value
// int meas2_adc()  setup Arduino ADC to do 2 conservative measurements
// int meas4_adc()  setup Arduino ADC to do 2 conservative measurements
// void meas328()   Setup Arduino ADC to do  measurements
// void meas1115() Setup ADS1115 A & B to do measurements
// void conv_float() convert all integer measurements to floating point values
// void self_chk()  Do safety check.
// void halt(int8_t fault) On error, shut the system dowm and halt
// void in_sw()  setup the input switch
// void balance() setup batteries balancing
// void plot_results()  Print out selected measurements for serial plot
// void measure_ind()  Base on measurement, calculate the effective inductance
// void prn_results()  Print out measurements to serial monitor
// void Lcd_prn()      Output measurements to LCD 
// void cal_Ah()       calculate Amp Hour for in, out and battery
// void buttonPressed()  interrupt routine to save measured data and pause the system. 
// void err_prn(int8_t fault)  print out reason of Error

LiquidCrystal_I2C lcd(0x27,20,4); //Set  LCD I2C address, column and row
// LCD reinitialize count
int lcd_cnt;

// PWM
#define PwmBaseFreq 20000    // base PWM frequency 31.25KHz
#define PwmMaxPw    248    // max PWM pulse width
#define PwmInc      1    // PWM pulse width increment

// Limits
#define lmt_Iout  10.0    // Output current limit
#define lmt_Iin  10.0     // Input current limit
#define lmt_temp  60.0    // Temperature limit
//Due to M1 body diode forward voltage, the internal input voltage can’t not be 0.8V lower than V12.  
#define lmt_intrV  0.8    //Internal input voltage limit
#define INOFF   12.2      // Input voltage threshold to shut off PWM
#define In_ON   12.0      // Input switch ON voltage threshold
#define battery_OV  12.8      // over voltage for combine batteries 
#define battery_UV  9.0      // under voltage for combine batteries 
#define out_UV  9.0      // under voltage for output

#define Max_unbal 0.2    // maximum unbalance threshold voltage 
#define Li_OV  4.3      // maximum Lithium battery OV limit
#define Li_UV  3.3      // minimum Lithium battery UD limit

#define Li_pw_thr  4.2    // Lithium Battery threshold. If exceeded pulse width decreases by 3

#define ainm  2.9298/4096.0  //Arduino ADC multiple factor from Vin
#define aoutm 2.6636/4096.0  //Arduino ADC multiple factor from Vout
#define at1m 1.0086/2048.0
#define at2m 1.0082/2048.0


#define v12M 3.246 // ADC multiple factor from v12
#define v8M 2.149  // ADC multiple factor from v8
#define v4M 1.08865  // ADC multiple factor from v4

#define ILM 2.5686// ADC multiplication factor for average inductor current 
//#define IinM 2.5339// ADC multiplication factor for average inductor current 
#define IinM 2.614// ADC multiplication factor for average inductor current 
#define IoutM 2.4578 //  ADC multiplication factor for output current  
#define VintrM 4.8441  //  ADC multiplication factor for internal input voltage 
#define V5M 1.9989  //  ADC multiplication factor Vcc, 5V

// Balancing 
#define MisThr    0.01    // Balancing threshold voltage

// ADS1115 setup
#define addr1115A 0x48
#define addr1115B 0x49
#define delay1115  100    // ADS1115 conversion delay in uS
#define RemeasT  10 // re-measure time due to sudden surge of current at output
#define display_cnt 100

// ADS1115 input selections 
uint8_t sel_A0=0b11000011;  // ADS1115A mode:A0-GND, 4.096v, single
uint8_t  sel_A1=0b11010011;  // ADS1115A mode:A1-GND, 4.096v, single
uint8_t   sel_A2=0b11100011;  // ADS1115A mode:A2-GND, 4.096v, single
uint8_t   sel_A3=0b11110011;  // ADS1115A mode:A3-GND, 4.096v, single

float ADC1b = 0.000125F;  // ADS1115 resolution at range of 4.096V.

struct measurement{
// measured by Arduino ADC
uint8_t  RadioId;  // NRF24L01 radio ID.
int16_t  ain;  // sum of 4 output voltages from A3
int16_t  aout;  // sum of 4 output voltages from A2
int16_t  at1;   // sum of 2 temperature 1 from A0
int16_t  at2;   // sum of 2 temperature 2 from A1


// Measured by ADS1115A
int16_t Iin;   //  input current
int16_t Li12;  // v12
int16_t Li8;   // v8
int16_t Li4;   // v4
// Measured by ADS1115B
int16_t IL;   // Input current

int16_t Iout;  // Output current
int16_t Vintr;   // Internal input voltage
int16_t V5;   // Vcc


int16_t PulseWidth;  // PWM pulse width
uint8_t    err;       // error code  0= no error. 
uint8_t  portb;//PORTB
uint8_t  portd; //PORTD  value
uint16_t SendCounts;  // number packages sent
};

long sum_IL;  // sum of inductor current in 100 measurements
long prv_IL;  // prvious sum of inductor current in 100 measurements
// measured by Arduino ADC
float  faout;  // output voltages
float  fat1;   // temperature 1
float  fat2;   // temperature 2

float pwr_in;  // input power
float pwr_out;  // output power
// float sum_pwr_in; // sum of 100 input power
// float avg_pwr_in,avg_prv_pwr; // average input power and prv_pwr
float    pwr_bat; // battery power  pos= charging  neg= discharging
float    i_bat;  // current flows into battery, pos= charging  neg= discharging

// Measured by ADS1115A
float fVin; // solar input voltage
float fLi12;  // v12
float fLi8;   // v8
float fLi4;   // v4
// Measured by ADS1115B
float fIL; // inductor  current
float fIin;  // input current
float fIout;  // output current
float fVintr;   // internal input voltage
float fV5;   // Vcc

float vb12,vb8,vb4;  //  voltage across battery
float vbmis;   //voltage difference between battery
// Variables to Calculate inductance
float inductance;   // Inductance
float period,on_time,fly_time; // Inductor switching time
float ipeak;  // peak current flows in the inductor
float ind_dv; // max voltage across switching inductor. 

float in_AS, out_AS, bat_AS; // Amp-SECOND for input, output, and  battery

int16_t pwm_dir;  //pulse width change direction 1= increment -1=decrement
float in_imp, out_imp;  // input and output switch impedance

int16_t  meas_cnt;  // count number of measurements before print result and update mode of operation

long t0,t1,t_prv,t_start,t_100s,t_100e,t_Ah;   // measurement time
// int16_t junk=0;  // temperatory storage
long freq = PwmBaseFreq;

float pwr_on_t;  // Total power on time in sec.
//******************************************************************

uint8_t meas2EEPROM;  //Press button Interrupt routine will set this to 1 inorder to save measurements into EEPROM
//******************************************************************

// Hardware configuration: Set up nRF24L01 radio on SPI bus (pins  11, 12, 13) plus pins 8 & 10
#define RF_CHAN 108  //Using Radio Channel
#define RADIO_TX_ID  13             // This Radio ID.
#define DESTINATION_RADIO_ID 18 // Transmit to Radio ID.
#define PIN_RADIO_CE  8
#define PIN_RADIO_CSN 10

NRFLite _radio;
measurement vm;
//******************************************************************
void setup() {
  int loop_cnt;
  Serial.begin(115200);  // set up  serial monitor baud rate

// PORTB bit 0 to 5 have no connection except bit 1, D9, which uses for PWM.
  DDRB |= 0b00111111; // set PORTB bit 0 to bit5 as output
  PORTB &= 0b11000000; // All PORTB out set to LOW.

// PORTD 
// bit 2: Falling edge Interrupt to save measurements to EEPROM
// bit 3: balance v8
// bit 4: balance v12
// bit 5: enable output switch
// bit 6: enable input switch
// bit 7: balance v4  low enable

  DDRD = (DDRD | 0b11111000) & 0b11111011; // set PORTD bit 3-7 as output bit 2 as input
  // disable input, output, and all balancing. bit 2 & 7 pull high, output switch enable
  PORTD =  (PORTD & 0b00000011) | 0b10100100; 
  delay(1000); // delay 1 sec for power up
  // Set I2C bus as a master 
  Wire.begin();  // initialize I2C
  scan_i2c();  // scan for I2C devices
// Initialize LCD display
  lcd.init();   //initialize LCD
  lcd.backlight();  // turn LCD light on  
  lcd.begin(20, 4); // set up the LCD's number of columns and rows:
  lcd.clear(); // clear LCD screen
  // read error data from EEPROM 
  EEPROM.get(0,vm);
  conv_float(); // convert all measurement to floating point
  prn_results();
  Lcd_prn();
  // read saved data from EEPROM 
  EEPROM.get(sizeof(vm),vm);
  conv_float(); // convert all measurement to floating point
  prn_results();
  Lcd_prn();
  
   // Initiate the NRF24L01
    if (!_radio.init(RADIO_TX_ID, PIN_RADIO_CE, PIN_RADIO_CSN, NRFLite::BITRATE250KBPS, RF_CHAN))
    {
        halt(12); // Stop the program.
    }

  detachInterrupt(digitalPinToInterrupt(2));  // disable interrupt from INT0

  Serial.println(F(" Press button to Continue"));
  Serial.println(F(" Subsequent Press, Measurements save to EEPROM"));
  lcd.setCursor(0,3);    // row 3 column 1.                                      
  lcd.print(F("Press Button to Cont"));  
    while ((PIND & 0b00000100) != 0); //wait for button to press
    while ((PIND & 0b00000100) == 0);    // wait for button released.

  vm.RadioId = RADIO_TX_ID;  // Over write the Radio ID saved in EEPROM
  vm.err = 0;  // clear error flag

// Set ADC prescaler  
//  ADCSRA = 0b10010110;   // set ADC clock prescaler to 64. Sample rate =16MHz/64 = 250KHz
  ADCSRA = 0b10010111;   // set ADC clock prescaler to 128. Sample rate = 16MHz/128 = 125KHz  
  
 // setup balancing PWM.
  freq = PwmBaseFreq;  // set to base frequency of 20KHz. 
  InitTimersSafe();    // setup PWM timer
  bool success = SetPinFrequencySafe(9,freq);
  if (success) {
//    vm.PulseWidth = PwmMaxPw/2;  // PWM pulse width set to half of max value
    vm.PulseWidth = 0;  // PWM pulse width set to 0
    pwmWrite(9,vm.PulseWidth);       //set PWM duty cycle  
  }  
  else halt(1);

  analogReference(DEFAULT);  // set ADC reference to Vcc
  delay(200);   // allow everythings have time to settle. 
  Serial.println(F("System Running "));
  meas(); // measure controller voltages

  self_chk();  //self diagonistic
  in_sw();     //set input switch
  pwm_dir=1; // set pulse width to increment direction. 1=increment -1=decrement
  meas_cnt=0;
  in_AS = 0.0;
  out_AS = 0.0;
  bat_AS =0.0;

  t_start = millis();
  t_prv=t_start;
  t_100s = t_start;
  t_Ah = t_start;
  pwr_on_t =0.0;
  vm.SendCounts = 0;
  sum_IL =0;  // sum of inductor current in 100 measurements
  prv_IL =0;  // previous sum of inductor current
  EEPROM.put(sizeof(vm),vm);  // save the initial measurement

EIFR = 0b11; // clear both interrupt flags, INTO and INT1
attachInterrupt(digitalPinToInterrupt(2),buttonPressed,FALLING);  // interrupt setup to 2, falling
}

//****************************************************************
void loop() {
  int i,j;
  t0 = millis();
  meas_cnt++;
  meas();
  self_chk();  
  t1 = millis();
   
  if (meas_cnt > display_cnt) {
    vm.SendCounts++;  // increament the print measure counts
//    avg_pwr_in = sum_pwr_in/100.0;
    in_sw();      // Enable or disable input switch
    pw_adj();     // Adjust Switching Pulse width
    balance();   // Setup which battery for balancing
    t_100e = millis();
    prn_results();  //Print measurement at Serial Monitor
    Lcd_prn();  //Display measurement at LCD
    // Send measurements out
    vm.portb=PORTB;
    vm.portd=PORTD;
    if (!_radio.send(DESTINATION_RADIO_ID, &vm, sizeof(vm), NRFLite::NO_ACK)) 
       halt(12);
    t_100s= t_100e;
    meas_cnt = 0;  
    prv_IL=sum_IL;
    sum_IL=0;
    }
}
//*****************************************************

//*****************************************************
//ISR function excutes when push button at pin D2 is pressed
void buttonPressed()    {  
   EIFR = 0b11; // clear both interrupt flat, INTO and INT1
   meas2EEPROM = 1;  // set save to EEPROM status. Measurement saved in prn_result() 
  }

//*****************************************************
 void cal_Ah() {
   long t_pass;   //time pass since previous calculation

  t_pass = t1-t_Ah;
  pwr_on_t +=t_pass;
  in_AS += 1.0e-3*t_pass * fIin;
  out_AS += 1.0e-3*t_pass * fIout;
  bat_AS += 1.0e-3*t_pass * i_bat;
  t_Ah = t1;
  
  // if 24 hours passed reset Ah and start time.
  if( (t1 - t_start) > 86400000L) {
    in_AS = 0.0;
    out_AS = 0.0;
    bat_AS =0.0;
    vm.SendCounts = 0;
    t_start = t1;  //reset start up time every 24 hours
   }
 }
 
//*****************************************************/
void pw_adj(){
// there is a switch on dead time of 1uS.
if(vm.PulseWidth < 5)  {
  vm.PulseWidth += PwmInc;
  pwm_dir = 1;
  PORTD =  (PORTD & 0b11100111) | 0b10000000; // disable all balancing
}
else {
    if (pwm_dir <0 ) {
     if ( sum_IL > prv_IL)  {
      vm.PulseWidth -= PwmInc; 
       pwm_dir = -1;
     }
     else {
      vm.PulseWidth += PwmInc ;
      vm.PulseWidth += PwmInc ;
      pwm_dir = 1;
     }
  }
  else {
     if (prv_IL > sum_IL)  {
      vm.PulseWidth -= PwmInc ;  
       pwm_dir = -1;     
     }
     else {
      vm.PulseWidth += PwmInc ;
      vm.PulseWidth += PwmInc ;
      pwm_dir = 1;
      }
    }
  }
  // check for over voltage
  if ((vb12 >  Li_pw_thr) ||(vb8 >  Li_pw_thr) || (vb4 >  Li_pw_thr) )  {
    vm.PulseWidth -= PwmInc;
    vm.PulseWidth -= PwmInc;
    vm.PulseWidth -= PwmInc;
  }
  if (vm.PulseWidth < 0) vm.PulseWidth = 0;
  if (vm.PulseWidth >PwmMaxPw) vm.PulseWidth =PwmMaxPw;
 // pwmWrite(9,133);
  pwmWrite(9,vm.PulseWidth);
}

/*****************************************************/
void meas(){

  meas328();    // do arduino ADC measurements
  meas1115();   // do ads1115 addr=0x48 & 0x49 measurements
  conv_float(); // convert all measurement to floating point
}

/******************************************************/
  // ADS1115 Configuration
  // MSB: Bits 15:8  ----------------------------------
  // Bit  15    1=Begin Single Conversion (power down mode)
  // Bits 14:12  AINP-AINN
  // 000 : A0-A1   001 : A0-A3   010 : A1-A3   011 : A2-A3
  // 100 : A0-GND  101 : A1-GND  110 : A2-GND  111 : A3-GND
  // Bits 11:9  +/-Range 
  // 000=6.144v  001=4.096v  010=2.048v  011=1.024  
  // 100=0.512v  101=0.256v  110=0.256v  111=0.256v
  // Bits 8     0=Continuous conversion, 1= single shot
  
  // LSB: Bits 7:0 ----------------------------------
  // Bits 7:5 Samples per second 
  // 000=8,   001=16,  010=32,  011=64,
  // 100=128, 101=250, 110=475, 111=860
  // Bit  4   Comparator Mode 0=Traditional, 1=Window
  // Bit  3   Comparator Polarity 0=low, 1=high
  // Bit  2   Latching 0=No, 1=Yes
  // Bits 1:0 Comparator # before Alert pin goes high
  //      00=1, 01=2, 10=4, 11=Disable this feature
 
// setup ads1115A operational mode
void config_ads1115a (  uint8_t msb) {
  // Step 1 
  Wire.beginTransmission(addr1115A); //prepare I2C data transfer 
  Wire.write(0b00000001);          // Point to Config Register
  Wire.write(msb); //
  Wire.write(0b11100011); //sample rate=860, Disable comparator
//  Wire.write(0b11000011); //sample rate=475, Disable comparator
//  Wire.write(0b10100011); //sample rate=250, Disable comparator
//  Wire.write(0b10000011); //sample rate=128, Disable comparator
//  Wire.write(0b01100011); //sample rate=64, Disable comparator
//  Wire.write(0b01000011); //sample rate=32, Disable comparator
  Wire.endTransmission();   // I2C WRITE to the module

  // Step 2 ====================================
  Wire.beginTransmission(addr1115A); //prepare I2C data transfer
  Wire.write(0b00000000);          //Point to Conversion register
  Wire.endTransmission();          // I2C WRITE to the module
}

//*****************************************************************
// setup ads1115B operational mode
void config_ads1115b (  uint8_t msb) {
  // Step 1 
  Wire.beginTransmission(addr1115B); //prepare I2C data transfer 
  Wire.write(0b00000001);          // Point to Config Register
  Wire.write(msb); //
  Wire.write(0b11100011); //sample rate=860, Disable comparator
  Wire.endTransmission();   // I2C WRITE to the module

  // Step 2 ====================================
  Wire.beginTransmission(addr1115B); //prepare I2C data transfer
  Wire.write(0b00000000);          //Point to Conversion register
  Wire.endTransmission();          // I2C WRITE to the module
}

//*******************************************************************
// Read ads1115A  register
int reg_ads1115a () {
  int adc_value;
  Wire.requestFrom(addr1115A, 2);  // Request 2 bytes 1st MSB then 2nd LSB
  adc_value = (Wire.read() << 8 | Wire.read());
  return (adc_value);
}

//***************************************************************************
// Read ads1115B  register
int reg_ads1115b () {
  int adc_value;
  Wire.requestFrom(addr1115B, 2);  // Request 2 bytes 1st MSB then 2nd LSB
  adc_value = (Wire.read() << 8 | Wire.read());
  return (adc_value);
}
//*************************************************************************
// read ads1115A value
int v_ads1115a() {
  int adc_value;
  Wire.beginTransmission(addr1115A); //prepare I2C data transfer
  Wire.write(0b00000000);          //Point to Conversion register
  Wire.endTransmission();          // I2C WRITE to the module
  adc_value = reg_ads1115a();
  return adc_value; 
}

//***********************************************************************
// read ads1115B value
int v_ads1115b() {
  int adc_value;
  Wire.beginTransmission(addr1115B); //prepare I2C data transfer
  Wire.write(0b00000000);          //Point to Conversion register
  Wire.endTransmission();          // I2C WRITE to the module
  adc_value = reg_ads1115b();
  return adc_value; 
}

//************************************************************************
// Trigger Arduino ADC 3 times, discard the 1st reading and sum up the last 2.
int meas2_adc() { 
 int value=0;
 
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); //wait for conversion is done
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); //wait for conversion is done}
  value += ADC;
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); //wait for conversion is done}
  value += ADC;
  return(value);
}

// Trigger Arduino ADC 5 times, discard the 1st reading and sum up the last 8.
int meas4_adc() { 
 int value=0;
 
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); //wait for conversion is done
  for(int i=0;i<4;i++) {
  ADCSRA |= _BV(ADSC); // Start conversion
  while (ADCSRA & 0b01000000); //wait for conversion is done}
  value += ADC;
  }
  return(value);
}

// measure all used arduino ADC. Discard 1st reading. 
void meas328(){
  int i;
     ADMUX = 0b01000000;  //Measure T1 at A0
     vm.at1 = meas2_adc();    
     ADMUX = 0b01000001;  //Measure T2 at A1
     vm.at2 = meas2_adc();         
     ADMUX = 0b01000010;  //Measure out at A2
     vm.aout = meas4_adc();    
     ADMUX = 0b01000011;  //Measure out at A3
     vm.ain = meas4_adc();  
  }

void meas1115(){
    config_ads1115a( sel_A3);   //measure Li4 voltage
    config_ads1115b( sel_A3);   //measure internal input voltage
    delayMicroseconds(delay1115);    // time for ADS1115 to convert,min 700uS for sample=860 
    vm.Li4 =v_ads1115a(); 
    vm.Vintr =v_ads1115b(); 

    config_ads1115a( sel_A2);   //measure Li8 voltage
    config_ads1115b( sel_A2);   //measure V5
    delayMicroseconds(delay1115);    // time for ADS1115 to convert,min 700uS for sample=860 
    vm.Li8 =v_ads1115a(); 
    vm.V5 =v_ads1115b(); 

    config_ads1115a( sel_A1);  //measure Li12 voltage
    config_ads1115b( sel_A1);  //measure output current
    delayMicroseconds(delay1115);    // time for ADS1115 to convert,min 700uS for sample=860 
    vm.Li12 =v_ads1115a(); 
    vm.Iout =v_ads1115b(); 

    config_ads1115a( sel_A0); //measure solar voltage
    config_ads1115b(sel_A0); //measure input current
    delayMicroseconds(delay1115);    // time for ADS1115 to convert,min 700uS for sample=860 
    vm.Iin =v_ads1115a(); 
    vm.IL =v_ads1115b(); 
  }

//******************************************************************
  // safety check
void self_chk(){
  uint8_t portdx;
  if ((fIout > lmt_Iout)&&(fIout < 2.0*lmt_Iout))  {
    delay(100);  // delay to avoid sudden surge of current 
    meas();   // redo the measurement
    conv_float(); 
    if (fIout > lmt_Iout) halt(2);  // Exceed output current limit  
   }

  if (fIin > lmt_Iin) halt(3);    // Exceed input current limit
  if ((fat1 > lmt_temp) || (fat2 > lmt_temp)) halt(4);  // temperature exceed limit
  if (fLi12 > battery_OV) halt(5);  // Exceed Li battery over all voltage limit
  if (fLi12 < battery_UV) halt(6);  // Exceed under voltage limit
  if ((vb12 > Li_OV)||(vb8 > Li_OV)||(vb4 > Li_OV)) { // Exceed individual OV
    delay(RemeasT);  // delay to avoid sudden surge of output current 
    meas();   // redo the measurement
    conv_float();   
    if ((vb12 > Li_OV)||(vb8 > Li_OV)||(vb4 > Li_OV)) halt(7); // Exceed individual OV
   }    
  if ((vb12 < Li_UV)||(vb8 < Li_UV)||(vb4 < Li_UV)) { // Exceed individual UV
    delay(RemeasT);  // delay to avoid sudden surge of output current 
    meas();   // redo the measurement
    conv_float();   
    if ((vb12 < Li_UV)||(vb8 < Li_UV)||(vb4 < Li_UV)) halt(8); // Exceed individual UV
   }    
  if (vbmis > Max_unbal) { // Exceed unbalance voltage limit
    delay(RemeasT);  // delay to avoid sudden surge of output current 
    meas();   // redo the measurement
    conv_float();   
    if (vbmis > Max_unbal) halt(9);  // Exceed unbalance voltage limit
   }    
  if (((PORTD & 0b00100000) != 0) && (faout < out_UV)){ 
    PORTD &= 0b11011111;  // turn the output switch off  

    delay(RemeasT);  // delay to avoid sudden surge of output current 
    PORTD |= 0b00100000;  // turn the output switch on      
    delay(RemeasT);  // delay to avoid sudden surge of output current 
 
    meas();   // redo the measurement
    conv_float();   
    if (((PORTD & 0b00100000) != 0) && (faout < out_UV))
       halt(10);  // Output voltage 
  }
  if ((PORTD & 0b00000000) && (fLi12-fVintr > lmt_intrV )) halt(11);  // Internal input drop exceed 0.8V 
}

//**********************************************
  void halt(int8_t fault) {
  vm.err = fault;
  // save port registers values
  vm.portb = PORTB;
  vm.portd = PORTD;
  EEPROM.put(sizeof(vm),vm);  // save the initial measurement
  PORTD =  (PORTD & 0b00000011) | 0b10000100; 
  PORTB &= 0b11000000; // All PORTB out set to LOW.
       prn_results();
       Lcd_prn();
       lcd.setCursor(12,3);                                          
       lcd.print(F(" PAUSED "));

  err_prn( fault);
  Serial.print(F(" System Stop on ERROR "));
  Serial.println(fault);
  while(1);
}
//******************************************************************
void err_prn(int8_t fault){
  switch (fault) {
    case 1:  Serial.print(F(" Fail to setup PWM at Pin 9 of frequency "));
             Serial.println(PwmBaseFreq);
             Serial.print(F("  Pulse Width  "));
             Serial.println(vm.PulseWidth);
             break;
    case 2:  Serial.print(F(" Output current exceed limit of ["));
             Serial.print(lmt_Iout);
             Serial.print(F("]  "));
             Serial.println(fIout,3);
             break;
    case 3:  Serial.print(F(" Input current exceed limit of ["));
             Serial.print(lmt_Iin);
             Serial.print(F("]  "));
             Serial.println(fIin,3);
             break;
    case 4:  Serial.print(F(" Temperature exceed limit of ["));
             Serial.print(lmt_temp);
             Serial.print(F("]  "));
             Serial.print(fat1,1);
             Serial.print(F("  "));
             Serial.println(fat2,1);
             break;
    case 5:  Serial.print(F(" Battery above OV of ["));
             Serial.print(battery_OV);
             Serial.print(F("]  "));
             Serial.println(fLi12,3);
             break;
    case 6:  Serial.print(F(" Battery  below UV ["));
             Serial.print(battery_UV);
             Serial.print(F("]  "));
             Serial.println(fLi12,3);
             break;
    case 7:  Serial.print(F(" Individual Cell OV ["));
             Serial.print(Li_OV);
             Serial.print(F("]  Vb12 "));
             Serial.println(vb12,3);
             Serial.print(F("  Vb8 "));
             Serial.println(vb8,3);
             Serial.print(F("  Vb4 "));
             Serial.println(vb4,3);
             break;
    case 8:  Serial.print(F(" Individual Cell UV ["));
             Serial.print(Li_UV);
             Serial.print(F("]  Vb12 "));
             Serial.println(vb12,3);
             Serial.print(F("  Vb8 "));
             Serial.println(vb8,3);
             Serial.print(F("  Vb4 "));
             Serial.println(vb4,3);
             break;
    case 9:  Serial.print(F(" Unbalance exceed limit of ["));
             Serial.print(Max_unbal);
             Serial.print(F("]  dif "));
             Serial.print(vbmis,3);
             Serial.print(F("  vb12 "));
             Serial.print(vb12,3);            
             Serial.print(F("  vb8 "));
             Serial.print(vb8,3);
             Serial.print(F("  vb4 "));
             Serial.println(vb4,3);
             break;
     case 10:  Serial.print(F(" Exceed Output UV of ["));
             Serial.print(out_UV);
             Serial.print(F("]  "));
             Serial.println(faout,3);
             break;
     case 11:  Serial.print(F(" Internal voltage limit below ["));
             Serial.print(lmt_intrV);
             Serial.print(F("]  fVin "));
             Serial.print(fVin,3);
             Serial.print(F("  fVintr "));
             Serial.println(fVintr,3);
             Serial.print(F("  fLi12 "));
             Serial.println(fLi12,3);
             break;      
     case 12:  Serial.println(F("Can't communicate with NRF24L01 "));                             
               break;
     case 13:  Serial.println(F("Manually Stop the System"));                             
               break;
    }
}
//******************************************************************
void in_sw(){
   // Input switch control
   if ((fVin> In_ON) && (fVin > fLi12))
      PORTD |=  0b01000000;   // Enable input switch, bit 6
   else {
      PORTD &=  0b10111111;   // Disable input switch, bit 6
      vm.PulseWidth=0;
      }
}

//******************************************************************
void balance(){
  if(( vbmis < MisThr)|| (vm.PulseWidth<5))   PORTD =  (PORTD & 0b11100111) | 0b10000000; // no balancing
  else {
    if ((vb4 < vb12) && (vb4 < vb8))PORTD= (PORTD & 0b01100111);  // balancing at vb4
    if ((vb8 < vb12) && (vb8 < vb4))PORTD= (PORTD & 0b01100111) | 0b10001000;  // balancing at vb8
    if ((vb12 < vb8) && (vb12 < vb4))PORTD=(PORTD & 0b01100111) | 0b10010000;  // balancing at vb12 
  }
}

//***********************************************************************
void conv_float(){  

  // convert all measurement to floating point
    fLi12 =v12M *ADC1b * vm.Li12;  // voltage at v12 reference to GND 
    fLi8 = v8M *ADC1b * vm.Li8;  // voltage at v8 reference to GND
    fLi4 = v4M *ADC1b * vm.Li4;  // voltage at v4 reference to GND
    fIL = ILM *ADC1b * vm.IL;  // Inductor  current
    sum_IL = sum_IL + vm.IL; // sum of Inductor current in 100 measurements
    fIin = IinM *ADC1b * vm.Iin;  // Input  current
   if (fIL < 0.0) fIL = 0.0;
    fIout = IoutM *ADC1b * vm.Iout;  // output current
    if (fIout < 0.0) fIout = 0.0;
    fVintr = VintrM *ADC1b * vm.Vintr;  // Input voltage internal 
    fV5 = V5M *ADC1b * vm.V5;    // Vcc=5V
//    fV5 = 4.9958;   // V5 fix due to fail in ADS1115, measured by DVM
    faout = aoutm*fV5*vm.aout;     // Output voltage
    fVin = ainm*fV5*vm.ain;     // Input voltage
  

    fat1 = at1m*fV5*vm.at1;        // temperature voltage at probe 1
    //convert voltage to temperature, equation is base on measurement.+/- 5C between 20C to 80C 
    if (fat1 < 3.455)  fat1 = 25.5*fat1 -17.5;
    else if (fat1 < 4.095) fat1 = 39.2*fat1 -64.1;
         else fat1 = 75.8*fat1 -215.2;

    fat2 = at2m*fV5*vm.at2;        // temperature voltage at probe 2
    //convert voltage to temperature, equation is base on measurement.+/- 5C between 20C to 80C 
    if (fat2 < 3.483)  fat2 = 25.2*fat2 -17.2;
    else if (fat2 < 4.143) fat2 = 37.9*fat2 -60.9;
         else fat2 = 921*fat2 -286.3; 
 
    vb12 = fLi12 - fLi8;  // v12 Li battery voltage
    vb8 = fLi8 - fLi4;    // v8 Li battery voltage
    vb4 = fLi4;           // v4 Li battery voltage

    vbmis = max(vb12, max(vb8,vb4)) - min(vb12, min(vb8,vb4));  //max difference in Li battery voltages 
    ind_dv = fIL-fLi12; // voltage across inductor
    i_bat = fIL -fIout;
    
    pwr_in = fVin*fIin;
    pwr_out = fLi12 * fIL;
    cal_Ah();  // calculate Amp Hour for input, output and battery
//    sum_pwr_in+=pwr_in;
}
//***********************************************************************

void measure_ind() {

  period = 1.0/freq;
  on_time = 1.0e-6 + period*vm.PulseWidth/256.0; // add 1uS for switch off delay
  ind_dv = (fVintr -fLi12) - 0.02*fIL; // minus the sense resistor 0.02 Ohms voltage drop
  ipeak = 2.0*fIin*period/on_time;
  inductance = on_time*ind_dv/ipeak;

   Serial.print(F("Period, uS\t"));  
   Serial.print(F("OnTime, uS\t"));
   Serial.print(F("ind_dv,V\t"));
   Serial.print(F("Ipeak, A\t"));
   Serial.println(F("Ind, uH"));

   Serial.print(1000000.0*period,3);
   Serial.print(F("\t\t"));  
   Serial.print(1000000.0*on_time,3);
   Serial.print(F("\t\t"));  
   Serial.print(ind_dv,3);
   Serial.print(F("\t\t"));  
   Serial.print(ipeak,3);
   Serial.print(F("\t\t"));  
   Serial.print(1e6*inductance,1);   
   Serial.println();
   }
//****************************************************************
//  print measurements
void prn_results() {

   Serial.print(F("Vin, V\t"));  
   Serial.print(F("\tV12, V\t"));
   Serial.print(F("\tV8, V\t"));
   Serial.print(F("\tV4, V\t")); 
   Serial.print(F("\tVout,V"));
   Serial.println();      
   Serial.print(fVin,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi12,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi8,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi4,3);
   Serial.print(F("\t\t"));  
   Serial.print(faout,3);
   Serial.println();
   
   Serial.print(F("Vintr, V"));
   Serial.print(F("\tVb12, V\t"));  
   Serial.print(F("\tVb8, V\t"));
   Serial.print(F("\tVb4, V\t"));
   Serial.print(F("\tvbmis, V"));
   Serial.println();
   Serial.print(fVintr,3);
   Serial.print(F("\t\t"));        
   Serial.print(vb12,3);
   Serial.print(F("\t\t"));  
   Serial.print(vb8,3);
   Serial.print(F("\t\t"));  
   Serial.print(vb4,3);
   Serial.print(F("\t\t"));  
   Serial.print(vbmis,3);
   Serial.println();
              
   Serial.print(F("Iin, A\t"));
   Serial.print(F("\tIL, A\t"));
   Serial.print(F("\tIout, A\t"));  
   Serial.print(F("\tI_bat, A"));
   Serial.print(F("\tV5, V"));
   Serial.println();

   Serial.print(fIin,3);
   Serial.print(F("\t\t"));  
   Serial.print(fIL,3);
   Serial.print(F("\t\t"));  
   Serial.print(fIout,3);
   Serial.print(F("\t\t")); 
   Serial.print(i_bat ,3);
   Serial.print(F("\t\t"));
   Serial.print(fV5,3);
   Serial.println();
             
   Serial.print(F("Freq Hz\t"));  
   Serial.print(F("\tPulse Width"));   
   Serial.print(F("\tTemp1, C\t"));  
   Serial.print(F("Temp2, C"));   
   Serial.print(F("\tPORTD"));  
   Serial.println(); 
   
   Serial.print(freq);
   Serial.print(F("\t\t"));  
   Serial.print(vm.PulseWidth);
   Serial.print(F("\t\t"));  
   Serial.print(fat1,1);
   Serial.print(F("\t\t"));  
   Serial.print(fat2,1);
   Serial.print(F("\t\t"));  
   Serial.print(PORTD,BIN);  
   Serial.println();  
   
   Serial.print(F("sum IL, W\t"));  
   Serial.print(F("Prv IL, W\t"));   
   Serial.print(F("Pwr Dir\t"));   
   Serial.print(F("\tPwr Out, W"));   
   Serial.print(F("\tEff, %"));   
   Serial.println(); 
   
   Serial.print(sum_IL);
   Serial.print(F("\t\t"));  
   Serial.print(prv_IL);
   Serial.print(F("\t\t"));  
   Serial.print(pwm_dir);
   Serial.print(F("\t\t"));  
   Serial.print(pwr_out,4);
   Serial.print(F("\t\t"));  
   Serial.print(100.0*pwr_out/pwr_in,1);
   Serial.println();     

   measure_ind();
   
   Serial.print(F("In AH\t"));  
   Serial.print(F("\tOut AH \t"));   
   Serial.print(F("\tBat AH\t")); 
   Serial.println();
   Serial.print(in_AS/3600.0,3);
   Serial.print(F("\t\t"));  
   Serial.print(out_AS/3600.0,3);
   Serial.print(F("\t\t"));  
   Serial.print(bat_AS/3600.0,3);
   Serial.println();
     
   Serial.print(F("PWR_ON_T, S\t"));
   Serial.print(display_cnt);
   
   Serial.print(F(" Measurements, mS"));  
   Serial.print(F("\t"));  
   Serial.print(F("\tERROR"));
   Serial.println();
   Serial.print(pwr_on_t/1000.0,1);
   Serial.print(F("\t\t"));  
   Serial.print(t_100e-t_100s);
   Serial.print(F("\t\t\t\t"));  
   Serial.println(vm.err);

   Serial.print(F("Measurement Counter " ));
   Serial.println(vm.SendCounts);
   Serial.println();
   if (meas2EEPROM == 1) {
    vm.portb = PORTB;
    vm.portd = PORTD;
    EEPROM.put(0,vm);  // save the initial measurement 
    meas2EEPROM = 0;   // clear the saving status
   }
}

//***********************************************************
void scan_i2c(){
 Serial.println("Scanning I2C");

  for (byte address = 1; address < 127; ++address) {
    // The i2c_scanner uses the return value of
    // the Write.endTransmisstion to see if
    // a device did acknowledge to the address.
    Wire.beginTransmission(address);
    byte error = Wire.endTransmission();

    if (error == 0) {
      Serial.print(F("Found I2C Addr 0x"));
      Serial.println(address, HEX);
      }
  }
}


//**********************************************************
// Print measurements to LCD display
void Lcd_prn()
   {
  uint8_t portd;
  static int a;
  float c_as,o_as;
// LCD display VChg, IChg, IOut, TC  
// re-initialize at every 30sec or 20 lcd counts so the LCD can swap out
  if (lcd_cnt > 20) {
  lcd.init();   //initialize LCD
  lcd.backlight();  // turn LCD light on  
  lcd.begin(20, 4); // set up the LCD's number of columns and rows:    lcd_cnt =0;
  lcd_cnt =0; 
  }
  else lcd_cnt++;


  lcd.clear();
  lcd.setCursor(0, 0);  // set the cursor to column 0, row 1
// 1st row, v12,v8,v4, * indicate balancing on that battery
  portd = PORTD;
  if((portd & 0b0010000)==0);
  else  lcd.print(F("*"));
  lcd.print(fLi12,3);  // v12  voltage, 6 chars
  if((portd & 0b00001000)==0)    lcd.print(F(" "));
  else  lcd.print(F(" *"));
  lcd.print(fLi8,3);  // v8  voltage, 5 chars 
  if((portd & 0b10000000)==0)    lcd.print(F(" *"));
  else  lcd.print(F(" "));
  lcd.print(fLi4,3);  // v4  voltage, 5 chars
//  lcd.print(F(" "));
//  lcd.print(1000*vbmis,0);  // Mismatch  voltage, 5 chars
//  lcd.print(F("m"));

  // row 2,  Vin, Vout, pulse width, PortD
  lcd.setCursor(0, 1);  // set the cursor to column 0, row 1
  lcd.print(fVin,2);  // vin  voltage
  lcd.print(F(" "));
  lcd.print(faout,2);  // vout  voltage
  lcd.print(F(" "));
  lcd.print(1000*vbmis,0);  // Mismatch  voltage
  lcd.print(F("m "));
  portd=PORTD;
  lcd.print(portd,HEX);  //  PORTD, 2 chars


  
  // row 2,  Vin, Vout, pulse width, PortD 
  lcd.setCursor(0, 2);  // set the cursor to column 0, row 1 
  lcd.print(fIL,1);   //charging current,5 chars
  lcd.print(F("A "));  // 
  lcd.print(fIout,1);   //output current, 5 chars
  lcd.print(F("A "));
  lcd.print(vm.PulseWidth);  // pulse width
  lcd.print(F(" "));
  lcd.print(fat1,0);  //Temperature 1, 4 chars
  lcd.print(F(" "));
  lcd.print(fat2,0);  //Temperature 1, 5 chars   
//  lcd.print(F("C"));

  lcd.setCursor(0, 3);  // set the cursor to column 0, row 3 
  if (vm.err > 0) {
  lcd.print(F("E"));
  lcd.print(vm.err);  //Error code, 1 chars   
  lcd.print(F(" B 0x"));
  lcd.print(vm.portb,HEX);  //Error code, 1 chars     
  lcd.print(F(" D 0x"));
  lcd.print(vm.portd);  //Error code, 1 chars   
  }
  else {
  lcd.print(in_AS/3600.0,1);  //input Amp Second
  lcd.print(F(" "));
  lcd.print(out_AS/3600.0,1);  //output Amp Second   
  lcd.print(F(" "));
  lcd.print(bat_AS/3600.0,1);  //battery Amp Second   
  lcd.print(F(" "));
  lcd.print(vm.SendCounts);  //battery Amp Second   
  }
 }
